import os.path
import sys

from docopt import docopt

from yfind.find import find_matching_files


def main():
    __doc__ = """Search YAML files satisfying specified conditions.

    Usage:
      yfind <search_exp> [<path>...]

    Options:
      -h --help     Show this screen.
    """

    arguments = docopt(__doc__, sys.argv[1:])
    paths = arguments.get('<path>', [])
    expression = arguments.get('<search_exp>')
    for match in find_matching_files(paths, expression):
        print(match)



import os

from yaml import load
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from yfind.parser import matches


def _file_matches_expression(filepath, expression):
    """Check if given file matches the expression.
    """
    with open(filepath) as f:
        parsed = load(f, Loader=Loader)
    return matches(parsed, expression)


def _find_matching_files_in_path(path, expression):
    """Recursively search given directory for files matching the expression.
    """
    matches = []
    for (dirpath, dirnames, filenames) in os.walk(path, followlinks=True):
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            if _file_matches_expression(filepath, expression):
                matches.append(filepath)
    return matches


def find_matching_files(paths, expression):
    """Search for files matching the search expression.

    ``paths`` is a list of paths to be searched recursively.
    ``expression`` is the expression to be satisfied by matching files.

    Returns a list of matching files.
    """

    matches = []
    for path in paths:
        matches += _find_matching_files_in_path(path, expression)
    return matches



class TestGrammar(object):
    def test_ops(self):
        from yfind.parser import matches
        data = {
            'foo': {'bar': 5},
            'bar': [1, 2, 3],
        }
        assert matches(data, ".foo.bar > .bar[2]")
        assert matches(data, "4 < .foo.bar")
        assert matches(data, ".bar[1] == 2")
        assert not matches(data, ".bar[1] != 2")
        assert matches(data, "not (.bar[1] != 2)")
        assert matches(data, ".bar[0] <= 1")
        assert matches(data, ".foo.bar >= 5")
        assert matches(data, "(.bar[1] == 2) and (.foo.bar == 5)")
        assert not matches(data, "(.bar[1] == 2) and (not (.foo.bar == 5))")
        assert matches(data, "(.bar[1] == 3) or (not (.foo.bar == 3))")
        assert matches(data, "?.bar[1]")
        assert not matches(data, "?.bar[5]")
        assert not matches(data, "?.void")

    def test_date(self):
        from yfind.parser import matches
        data = {
            'foo': "2008/11/5",
        }
        assert matches(data, ".foo == 2008/11/05")
        assert matches(data, ".foo > 2007/10/15")
        assert matches(data, ".foo < 2008/12/04")



from datetime import date
from functools import reduce
import operator as op

from modgrammar import EOF
from modgrammar import Grammar
from modgrammar import L
from modgrammar import WORD
from modgrammar import LIST_OF
from modgrammar import OPTIONAL
from modgrammar import ParseError
from modgrammar import REF


class Operator(Grammar):
    op_map = {
        '==': op.eq,
        '<': op.lt,
        '>': op.gt,
        '>=': op.ge,
        '<=': op.le,
        '!=': op.ne,
    }
    grammar = reduce(op.or_, [L(item) for item in op_map.keys()])

    def satisfied_by(self, left, right):
        op = self.op_map[self.elements[0].string]
        return op(left, right)


class Year(Grammar):
    grammar = WORD("0-9", count=4)


class Month(Grammar):
    grammar = WORD("0-9", min=1, max=2)


class Day(Grammar):
    grammar = WORD("0-9", min=1, max=2)


class Date(Grammar):
    grammar_whitespace = False
    grammar = (Year, '/', Month, '/', Day)

    def value(self, data=None):
        return date(int(self.elements[0].string),
                    int(self.elements[2].string),
                    int(self.elements[4].string))


class Scalar(Grammar):
    grammar_whitespace = False
    grammar = ('"', WORD(r'^\"'), '"')

    def value(self, data=None):
        return self.elements[1].string


class Floating(Grammar):
    grammar_whitespace = False
    grammar = WORD("0-9.")

    def value(self, data=None):
        return float(self.elements[0].string)


class Integral(Grammar):
    grammar_whitespace = False
    grammar = WORD("0-9")

    def value(self, data=None):
        return int(self.elements[0].string)


class FieldValue(Grammar):
    grammar = (Date | Floating | Integral)

    def value(self, data=None):
        return self.elements[0].value(None)


class Identifier(Grammar):
    grammar_whitespace = False
    grammar = (WORD("A-Za-z_"), OPTIONAL(L("["), WORD("0-9"), L("]")))


class Node(Grammar):
    grammar_whitespace = False
    grammar = (L('.'), LIST_OF(Identifier, sep='.', min=1))

    def _parse_value(self, exp):
        if isinstance(exp, str) or isinstance(exp, bytes):
            try:
                operand = FieldValue.parser().parse_string(exp, eof=True)
                return operand.value()
            except ParseError:
                pass
        return exp

    def _traverse(self, data):
        pos = data
        for element in self.elements[1]:
            if element.string == '.':
                continue
            if pos.get(element[0].string, None) is None:
                return None
            pos = pos[element[0].string]
            if element[1] is not None:
                index = int(element[1][1].string)
                if index < len(pos):
                    pos = pos[index]
                else:
                    return None
        return pos

    def exists(self, data):
        return self._traverse(data) is not None

    def value(self, data):
        pos = self._traverse(data)
        return pos is not None and self._parse_value(pos) or None


class Operand(Grammar):
    grammar = (Node | Date | Scalar | Floating | Integral)

    def value(self, data):
        return self.elements[0].value(data)


class BooleanUnaryOperator(Grammar):
    grammar = (L("not"))

    def bool_value(self, right, data):
        return not right.bool_value(data)


class BooleanBinaryOperator(Grammar):
    op_map = {
        'and': op.and_,
        'or': op.or_,
    }
    grammar = reduce(op.or_, [L(item) for item in op_map.keys()])

    def bool_value(self, left, right, data):
        op = self.op_map[self.elements[0].string]
        return op(left.bool_value(data), right.bool_value(data))


class Comparision(Grammar):
    grammar = (Operand, Operator, Operand)

    def bool_value(self, data):
        (left, op, right) = self.elements
        if left.value(data) is None or right.value(data) is None:
            return False
        return op.satisfied_by(left.value(data), right.value(data))


class BooleanUnaryComparision(Grammar):
    grammar = (BooleanUnaryOperator, '(', REF('BooleanExpression'), ')')

    def bool_value(self, data):
        (op, right) = self.elements[0], self.elements[2]
        return op.bool_value(right, data)


class BooleanBinaryComparision(Grammar):
    grammar = ('(', REF('BooleanExpression'), ')',
               BooleanBinaryOperator, '(', REF('BooleanExpression'), ')')

    def bool_value(self, data):
        left, op, right = self.elements[1], self.elements[3], self.elements[5]
        return op.bool_value(left, right, data)


class Existance(Grammar):
    grammar = ('?', Node)

    def bool_value(self, data):
        return self.elements[1].exists(data)


class BooleanExpression(Grammar):
    grammar = (
        Comparision
        | Existance
        | BooleanUnaryComparision
        | BooleanBinaryComparision
    )

    def bool_value(self, data):
        return self.elements[0].bool_value(data)


class SearchGrammar(Grammar):
    """Grammar for search expressions."""
    grammar = (BooleanExpression, EOF)

    @staticmethod
    def matches(data, exp):
        tree = SearchGrammar.parser().parse_string(exp, eof=True)
        return tree.elements[0].bool_value(data)


def matches(data, expression):
    """Check if data satisfies the expression.
    """
    return SearchGrammar.matches(data, expression)



import os

from setuptools import setup
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
    CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()
except IOError:
    README = CHANGES = ''

install_requires = [
    'docopt',
    'modgrammar',
    'pyyaml',
    ]

tests_require = [
    'mock',
    'pytest',
    'pytest-cov',
    'pytest-pep8',
    'pytest-xdist',
    ]

setup(name='yfind',
      version='0.1.0a2',
      description="Search YAML files satisfying specified conditions.",
      long_description='\n\n'.join([README, CHANGES]),
      classifiers=[
          "Programming Language :: Python",
          "Development Status :: 3 - Alpha",
          "Environment :: Console",
          "License :: OSI Approved :: BSD License",
          "Programming Language :: Python",
          "Programming Language :: Python :: 3",
          "Topic :: Text Processing",
          "Topic :: Text Processing :: Markup",
      ],
      author='Christian Neumann',
      author_email='cneumann@datenkarussell.de',
      url='https://github.com/chrneumann/yfind',
      keywords='yaml find search',
      license="BSD-2-Clause",
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=install_requires,
      tests_require=tests_require,
      entry_points="""\
      [console_scripts]
      yfind = yfind:main
      """,
      extras_require={
          'testing': tests_require,
          },
      )
